/*
 	Drivers_EL_AZ_local.c
	
	Drivers for local control of Elevation (EL) and Azimuth (AZ)
 	
	
	
	Copyright 2007 Daniel Berg
	
 	This file is part of SatFinder.
 
    SatFinder is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
 
    Foobar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SatFinder.  If not, see <http://www.gnu.org/licenses/>.
	
	*/
	

#include <avr/io.h>
#include <inttypes.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/eeprom.h>
	
	
#define EL_DIR_RELAY 	2			// On portD
#define EL_ON_RELAY 	3
#define AZ_PULSES_IN 	4
#define EL_PULSES_IN 	5
#define AZ_DIR_RELAY 	6			
#define AZ_ON_RELAY 	7			
	
#define UP 		1
#define DOWN 	0
#define RIGHT	1
#define LEFT 	0	
	
/* Stopptid fr stlldonet (tid det tar fr stlldonet att stanna frn
   det att strmmen stngts av */
#define STOPPTID 0.7		// sekunder
#define STOPPTID_M STOPPTID/0.032768	
	

// globala variabler	
volatile uint16_t el_pulses;
volatile uint8_t el_direction;
volatile uint16_t az_pulses;
volatile uint8_t az_direction;
volatile uint16_t el_target;
volatile uint16_t az_target;
volatile uint16_t az_min;
volatile uint16_t az_max;
volatile uint16_t el_min;
volatile uint16_t el_max;
volatile uint16_t az_zero;
volatile uint16_t az_360;
volatile uint16_t el_zero;
volatile uint16_t el_90;

volatile uint8_t el_in_motion;
volatile uint8_t az_in_motion;

volatile uint8_t Stoppfaktor_el;
volatile uint8_t Stoppfaktor_az;
volatile uint8_t TCNT0H;
volatile uint8_t OCR0A_high;
volatile uint8_t latest_stop_pulses_up;
volatile uint8_t latest_stop_pulses_down;
volatile uint8_t latest_stop_pulses_az;
volatile uint16_t el_pulses_old;
volatile uint16_t az_pulses_old;

// Dessa sparas i EEPROM:
uint16_t el_pulses_ee EEMEM;
uint16_t az_pulses_ee EEMEM;
uint16_t az_min_ee EEMEM;
uint16_t az_max_ee EEMEM;
uint16_t el_min_ee EEMEM;
uint16_t el_max_ee EEMEM;
uint16_t az_zero_ee EEMEM;
uint16_t az_360_ee EEMEM;
uint16_t el_zero_ee EEMEM;
uint16_t el_90_ee EEMEM;




// Lokala prototyper 
void EL_stop(void);
void AZ_stop(void);





uint8_t EL_init_driver (void) {
	el_pulses = eeprom_read_word(&el_pulses_ee);
//	el_target = el_pulses;
	el_min = eeprom_read_word(&el_min_ee);
	el_zero = eeprom_read_word(&el_zero_ee);
	el_90 = eeprom_read_word(&el_90_ee);
	el_max = eeprom_read_word(&el_max_ee);
	el_in_motion = 0;
	latest_stop_pulses_up = 0;
	latest_stop_pulses_down = 0;
	
	// initiera utgngar (samma rad i bde EL och AZ initieringen)
	DDRD = (_BV(EL_DIR_RELAY) | _BV(EL_ON_RELAY) | _BV(AZ_DIR_RELAY) | _BV(AZ_ON_RELAY));
	
	return 0;		//allt ok.
}

/**
	Returnera en gonblicksposition
*/
uint16_t EL_get_position (void) {
uint16_t el_total_pulses;
	// Kolla om timer1 r igng, i s fall mste rkna ut ett gonblicksvrde
	if ((Stoppfaktor_el) || (TCCR1B != 0)) {
		/* Rkna ut det totala aktuella antalet pulser, bde bokfrda och nyligen tillkommna */
		if (el_direction == UP) el_total_pulses = el_pulses + TCNT1;
		else el_total_pulses = el_pulses - TCNT1;
	} 
	else el_total_pulses = el_pulses;
	
	return el_total_pulses;
}

uint16_t EL_get_min (void) {
	return el_min;
}

uint16_t EL_get_max (void) {
	return el_max;
}

uint16_t EL_get_low (void) {
	return el_zero;
}

uint16_t EL_get_high (void) {
	return el_90;
}

/**
	Returnerar 0 om rotorn r stilla
*/
uint8_t EL_in_motion (void) {
	return el_in_motion;
}

uint16_t EL_get_status (uint8_t asked_status) {
	return 0;
}


/**
	Stt igng rotorn och kr till nskad position
	Positionen r ett 16 bitars heltal mellan min och max.
	Allts inte ett gradtal.
	*/
uint8_t EL_set_position (uint16_t new_position) {

	/* Ta hand om positioner som r utanfr vra grnser */
	if (new_position > el_max) new_position = el_max;
	if (new_position < el_min) new_position = el_min;
	
	// Frst en koll ifall vi verhuvudtaget behver gra ngot:
	if (new_position != el_pulses) {		
		
		// Kolla om EL redan r i rrelse, i s fall mste vi frst stoppa
		if (EL_in_motion()) {
			EL_stop();
			while (EL_in_motion());
		}
		
		if (new_position > el_pulses) {
			OCR1A = new_position - el_pulses - latest_stop_pulses_up;	// stt nytt Output Compare-ml 
			TIMSK1 = _BV(OCIE1A);						// aktivera avbrott fr output compare!
			el_direction = UP;							// Fr att mjukvaran skall rkna t rtt hll
			PORTD |= _BV(EL_DIR_RELAY);					// Justera riktning p stlldonet
		}
		else  {
			OCR1A = el_pulses - new_position - latest_stop_pulses_down;	// stt nytt Output Compare-ml 
			TIMSK1 = _BV(OCIE1A);						// aktivera avbrott fr output compare!
			el_direction = DOWN;
			PORTD &= ~(_BV(EL_DIR_RELAY));
		}
		TCCR1B = _BV(CS12) | _BV(CS11);				// aktivera rknaren
		_delay_ms(15);									// vnta lngre n ndvndigt p att 
		_delay_ms(15);									//   riktningsrelet skall sl om
		el_in_motion = 1;
		PORTD |= _BV(EL_ON_RELAY);						// aktivera stlldonsmotor
	}
	return 0;
}

/*
	Starta en kalibreringsprocedur fr EL.
*/	
uint16_t EL_cal_start (void) {
	el_pulses_old = el_pulses;			// kanske kan vi terta vr position?
	el_pulses = 12000;
	el_min = 0;
	el_max = 0xFFFF;
	el_zero = 0;
	el_90 = 0;
	return el_pulses;
}

/*
	Skriv kalibreringen till EEPROM
	Returnerar 0 om allt gick vl
	Returnerar 1 om nt gick fel, alla vrden terstllda/ofrndrade
*/
uint8_t EL_cal_write (uint16_t min, uint16_t low, uint16_t high, uint16_t max) {
uint8_t ret_val = 0;
	/* Kontrollera data mot villkor. Om ogiltigt anvnder vi gamla vrden */
	if (((high - low) < 90) ||
		(min > low) ||
		(max < high)) {
		/* terstll */
		EL_init_driver();
		el_pulses = el_pulses_old + ((int16_t) el_pulses - 12000);		// el_pulses kan bli galet hr
		
		ret_val = 1;
	} else {
		/* Allt okej, spara ny kalibrering i EEPROM */
		eeprom_write_word(&el_min_ee, min);
		eeprom_write_word(&el_max_ee, max);
		eeprom_write_word(&el_zero_ee, low);
		eeprom_write_word(&el_90_ee, high);
		/* Spara ven i RAM */
		el_min = min;
		el_max = max;
		el_zero = low;
		el_90 = high;
	}
	
	return ret_val;
}


uint8_t EL_set_config (uint16_t req_conf, uint16_t new_config_data) {
	return 0;	// 0 = allt ok.
}

/*
	Stoppa stlldonsmotorn och aktivera timer fr fysisk stopptid.
	*/
void EL_stop(void) {
	/* Kr frdrjningar och Stopptid enbart om drivningen verkligen varit igng.
	   D slipper vi frdrjningar i ondan om vi kallat p el_stop() som skerhetstgrd. */
	   
	if ((PIND & _BV(EL_ON_RELAY))) { 
		TIMSK1 &= ~(_BV(OCIE1A));			// undvik fler avbrott
		
		PORTD &= ~(_BV(EL_ON_RELAY) ); 		
		_delay_ms(15);						// vnta s att frihjulsdioden hinner ta strmmspiken
		PORTD &= ~(_BV(EL_DIR_RELAY));		// slpp riktningsrelet
		
		/* aktivera Timer2 fr att automatiskt stnga av rknaren efter den uppskattade
		   stopptiden fr stlldonet.
		   */
		TCNT2 = 0;
		TCCR2B = _BV(CS12) | _BV(CS11) | _BV(CS10);
		TIMSK2 = _BV(TOIE2);
		Stoppfaktor_el = STOPPTID_M;
	}
}


/**********************************************************
	AZ-avdelning
*/
uint8_t AZ_init_driver (void) {
	az_pulses = eeprom_read_word(&az_pulses_ee);
//	az_target = az_pulses;
	az_min = eeprom_read_word(&az_min_ee);
	az_zero = eeprom_read_word(&az_zero_ee);
	az_360 = eeprom_read_word(&az_360_ee);
	az_max = eeprom_read_word(&az_max_ee);
	az_in_motion = 0;
	latest_stop_pulses_az = 0;
	
	// initiera utgngar (samma rad i bde EL och AZ initieringen)
	DDRD = (_BV(EL_DIR_RELAY) | _BV(EL_ON_RELAY) | _BV(AZ_DIR_RELAY) | _BV(AZ_ON_RELAY));
	
	return 0;
}

uint16_t AZ_get_position (void) {
uint16_t az_total_pulses;
	// Kolla om timer0 r igng, i s fall mste rkna ut ett gonblicksvrde
	if ((Stoppfaktor_az) || (TCCR0B != 0)) {
		cli();		// vi lser in ett 16-bits tal som manipuleras i ett avbrott, stng av globala avbrott
		/* Rkna ut det totala aktuella antalet pulser, bde bokfrda och nyligen tillkommna */
		if (az_direction == RIGHT) {		// justera den officiella rkningen
			az_total_pulses = az_pulses + (uint16_t) TCNT0;				// addera lga 8 bitar
			az_total_pulses += (TCNT0H * 256);	// addera hga 8 bitar
		}
		else {
			az_total_pulses = az_pulses - (uint16_t) TCNT0;				// addera lga 8 bitar
			az_total_pulses -= (TCNT0H * 256);	
		}
		sei();		// tillt globala avbrott igen, senare borde vi kanske kolla ifall flaggan var satt frn brjan...
	} else az_total_pulses = az_pulses;
	
	return az_total_pulses;
}	
	
uint16_t AZ_get_min (void) {
	return az_min;
}

uint16_t AZ_get_max (void) {
	return az_max;
}

uint16_t AZ_get_low (void) {
	return az_zero;
}

uint16_t AZ_get_high (void) {
	return az_360;
}

/**
	Returnerar 0 om rotorn r stilla
*/
uint8_t AZ_in_motion (void) {
	return el_in_motion;
}

uint16_t AZ_get_status (uint8_t asked_status) {
	return 0;
}

/*
	Starta en kalibreringsprocedur fr AZ.
*/	
uint16_t AZ_cal_start (void) {
	az_pulses_old = az_pulses;			// kanske kan vi terta vr position?
	az_pulses = 12000;
	az_min = 0;
	az_max = 0xFFFF;
	az_zero = 0;
	az_360 = 0;
	return az_pulses;
}

/*
	Skriv kalibreringen till EEPROM
	Returnerar 0 om allt gick vl
	Returnerar 1 om nt gick fel, alla vrden terstllda/ofrndrade
*/
uint8_t AZ_cal_write (uint16_t min, uint16_t low, uint16_t high, uint16_t max) {
uint8_t ret_val = 0;
	/* Kontrollera data mot villkor. Om ogiltigt anvnder vi gamla vrden */
	if (((high - low) < 90) ||
		(min > low) ||
		(max < high)) {
		/* terstll */
		AZ_init_driver();
		az_pulses = az_pulses_old + ((int16_t) az_pulses - 12000);		// az_pulses kan bli galet hr
		
		ret_val = 1;
	} else {
		/* Allt okej, spara ny kalibrering i EEPROM */
		eeprom_write_word(&az_min_ee, min);
		eeprom_write_word(&az_max_ee, max);
		eeprom_write_word(&az_zero_ee, low);
		eeprom_write_word(&az_360_ee, high);
		/* Spara ven i RAM */
		az_min = min;
		az_max = max;
		az_zero = low;
		az_360 = high;
	}
	
	return ret_val;
}

uint8_t AZ_set_position (uint16_t new_position) {
uint16_t temp;

	/* Ta hand om positioner som r utanfr vra grnser */
	if (new_position > az_max) new_position = az_max;
	if (new_position < az_min) new_position = az_min;
	
	if (new_position != az_pulses) {
		/* Om AZ redan r igng s mste vi stanna frst */	
		if (AZ_in_motion()) {
			AZ_stop();
			while (AZ_in_motion());
		}
		
		if (new_position > az_pulses) {
			temp = new_position - az_pulses - latest_stop_pulses_az;
			OCR0A = (uint8_t) temp;						// stt nytt Output Compare-ml 
			OCR0A_high = (uint8_t) (temp >> 8);			// OCR0A r bara 8 bitar, vi frlnger manuellt
			az_direction = RIGHT;
			PORTD |= _BV(AZ_DIR_RELAY);
		}
		else {
			temp = az_pulses - new_position - latest_stop_pulses_az;
			OCR0A = (uint8_t) temp;						// stt nytt Output Compare-ml
			OCR0A_high = (uint8_t) (temp >> 8);			// OCR0A r bara 8 bitar, vi frlnger manuellt
			az_direction = LEFT;
			PORTD &= ~(_BV(AZ_DIR_RELAY));
		}
		TCCR0B = _BV(CS02) | _BV(CS01);
		TIMSK0 = _BV(OCIE0A) | _BV(TOIE0);						// aktivera ven avbrott vid overflow!
		_delay_ms(15);
		_delay_ms(15);
		az_in_motion = 1;
		PORTD |= _BV(AZ_ON_RELAY);
	}
	return 0;
}



uint8_t AZ_set_config (uint16_t req_conf, uint16_t new_config_data) {
	return 0; 		// 0 = allt vl.
}	

/*
	Stoppa azimuth-motorn och aktivera timer fr fysisk stopptid.
	*/
void AZ_stop(void) {
	/* Kr frdrjningar och Stopptid enbart om drivningen verkligen varit igng.
	   D slipper vi frdrjningar i ondan om vi kallat p az_stop() som skerhetstgrd. */
	
	if ((PIND & _BV(AZ_ON_RELAY))) {
		TIMSK0 &= ~(_BV(OCIE0A));			// undvik fler avbrott
		
		PORTD &= ~(_BV(AZ_ON_RELAY));			
		_delay_ms(15);
		PORTD &= ~(_BV(AZ_DIR_RELAY));
		
		/* aktivera Timer2 fr att automatiskt stnga av rknaren efter den uppskattade
		   stopptiden fr stlldonet.
		   */
		TCNT2 = 0;
		TCCR2B = _BV(CS12) | _BV(CS11) | _BV(CS10);
		TIMSK2 = _BV(TOIE2);
		Stoppfaktor_az = STOPPTID_M;
	}
}




/**************************************************************
	Avdelning som hanterar avbrott
*/
/**
	- Elevation -
	Vi har ntt vrt brvrde. Stanna rotorn.
	*/
ISR (TIMER1_COMPA_vect)
{
	EL_stop();							// stoppa maskineriet
}	

/**
	- Azimuth -
	Vi har ntt vrt brvrde. Stanna rotorn.
	*/
ISR (TIMER0_COMPA_vect)
{
	if (OCR0A_high == TCNT0H) {				
		AZ_stop();							// stoppa maskineriet
		TIMSK0 &= ~(_BV(OCIE0A));			// undvik fler avbrott
	}
}	

/**
	- Azimuth -
	Utka Timer0 med 8 bitar.
	*/
ISR (TIMER0_OVF_vect)
{
	TCNT0H++;
}	

/**
	Timeout fr timer2. 
	Stnger av rknaren T1. (Skall p ngot stt ven hantera T0... n'n gng...)
	*/
ISR (TIMER2_OVF_vect)
{
	if (Stoppfaktor_el) {
		if (--Stoppfaktor_el == 0) {
			// Addera det antal pulser som vi krt.
			if (el_direction == UP) {
				el_pulses += TCNT1;		// Justera den officiella rkningen
				// Hll koll p hur lngt frbi mlet vi krt
				if (TCNT1 > OCR1A) latest_stop_pulses_up = (uint8_t) TCNT1 - OCR1A;
				else latest_stop_pulses_up = 0;
			}
			else {
				el_pulses -= TCNT1;
				// Hll koll p hur lngt frbi mlet vi krt
				if (TCNT1 > OCR1A) latest_stop_pulses_down = (uint8_t) TCNT1 - OCR1A;
				else latest_stop_pulses_down = 0;
			}
			
			TCNT1 = 0;							// nolla rknaren
			el_in_motion = 0;
			eeprom_write_word(&el_pulses_ee, el_pulses);
		}
	}
	
	if (Stoppfaktor_az) {
		if (--Stoppfaktor_az == 0) {
			// Addera det antal pulser som vi krt.
			if (az_direction == RIGHT) {		// justera den officiella rkningen
				az_pulses += TCNT0;				// addera lga 8 bitar
				az_pulses += (TCNT0H * 256);	// addera hga 8 bitar
			}
			else {
				az_pulses -= TCNT0;				// samma med minus...
				az_pulses -= (TCNT0H * 256);
			}
			// Hll koll p hur lngt frbi mlet vi krt
			if (TCNT0 > OCR0A) latest_stop_pulses_az = (uint8_t) TCNT0 - OCR0A;
				else latest_stop_pulses_az = 0;
				
			TCNT0 = 0;							// nolla rknaren
			TCNT0H = 0;
			
			az_in_motion = 0;
			eeprom_write_word(&az_pulses_ee, az_pulses);
		}
	}
	
	if ((Stoppfaktor_el == 0) && (Stoppfaktor_az == 0)) {		
		TCCR1B = 0;					// Frhindra brus i rknaren.
		TCCR0B = 0;
		TCCR2B = 0;					// stng av timer2 ocks. Den har nu gjort sitt.
		TIMSK2 = 0;					// Inga fler avbrott hr heller (egentligen ondigt)
	}
}		
	


