/******************************* C SOURCE FILE ******************************
**
** Project		: LCD Library (4-bit Mode)
** Filename		: lcd.c
** Chip Type	: Any AVR
** Clock MHz	: 
** Compiler		: AVR GCC
**
*****************************************************************************
** 
** Copyright (c) 2006, Andrew Nickson, Daniel Berg 2007
** All rights reserved.
**
** You are free to use this code as you see fit, a mention would be nice and any
** feedback.
** Email: andy@re-mastered.co.uk 
** Web:   www.re-mastered.co.uk
**
*****************************************************************************
** VERSION HISTORY:
** ----------------
** Senare version av Daniel Berg
** 
** Version : 1.00
** Created by : Andrew Nickson		Date :  8th January 2006
** Description :  
**			<> 	Initial Basic Library
**			<>	Goto now allows any value to be entered.
**			<>	Added Options to output int and hex values.
**			
****************************************************************************/

#include "avr_lcd.h"


/* Local prototypes */
void lcd_busy(void);
void lcd_enable(void);

extern inline void LCD_PORT_MOD (uint8_t data) {
	LCD_PORT = (data<<2);
}

/***************************************************
 *  Display initialization
 ***************************************************/
void lcd_init(void)
{
  	LCD_DDR = 0xff;	//Make the LCD port an output
	
	_delay_ms(20);	//Allow LCD to power up > 15mS
  
  	//LCD Starts up in 8-Bit mode, busy cannot be checked at moment
  	//we send these bytes according to the datasheet.
  	LCD_PORT_MOD(0x03); lcd_enable();  _delay_ms(1);
  	LCD_PORT_MOD(0x03); lcd_enable();  _delay_ms(1);
  	LCD_PORT_MOD(0x03); lcd_enable();
  
  	//Now we can check the Busy
  	lcd_busy();
  
  	LCD_PORT_MOD(0x02); lcd_enable();	//This places the LCD in 4-bit mode.
  
  	lcd_busy();
  	LCD_PORT_MOD(0x02); lcd_enable();	//FUNCTION SET
  	LCD_PORT_MOD(0x08); lcd_enable();	
  	lcd_busy();
  	LCD_PORT_MOD(0x00); lcd_enable();	//DISPLAY/CURSOR
  	LCD_PORT_MOD(0x0C); lcd_enable();
	lcd_busy();
  	LCD_PORT_MOD(0x00); lcd_enable();	//ENTRY SET - Increment, No Disp Shift
  	LCD_PORT_MOD(0x06); lcd_enable();

  	lcd_cls();	//Clear Screen
	lcd_home();	//Home Cursor
}

/***************************************************
 *  Check if the LCD is Busy
 ***************************************************/
void lcd_busy(void)
{
	//We should check the busy flag properly, but 
	//were going to cheat and just use a delay to
	//allow the LCD to not be busy
	_delay_ms(2);
}

/***************************************************
 *  Toggle the Enable Pin on the LCD
 ***************************************************/
void lcd_enable(void)
{
	SET_E;		//E=1
	_delay_us(0.5);
	CLEAR_E;	//E=0
}


/***************************************************
 *  Clear Display
 ***************************************************/
void lcd_cls(void)
{
  	lcd_busy();
  	LCD_PORT_MOD(0x00); lcd_enable();	//CLEAR SCREEN
  	LCD_PORT_MOD(0x01); lcd_enable();
}

/***************************************************
 *  Go to Home Position (Top Left)
 ***************************************************/
void lcd_home(void)
{
  	lcd_busy();
  	LCD_PORT_MOD(0x00); lcd_enable();
  	LCD_PORT_MOD(0x02); lcd_enable();
}
/***************************************************
 *  Turn on cursor (blinking)
 ***************************************************/
void lcd_cursor_on(void)
{
  	lcd_busy();
  	LCD_PORT_MOD(0x00); lcd_enable();
  	LCD_PORT_MOD(0x0F); lcd_enable();
}
/***************************************************
 *  Turn off cursor
 ***************************************************/
void lcd_cursor_off(void)
{
  	lcd_busy();
  	LCD_PORT_MOD(0x00); lcd_enable();
  	LCD_PORT_MOD(0x0C); lcd_enable();
}
/***************************************************
 *  Turn on display
 ***************************************************/
void lcd_disp_on(void)
{
  	lcd_busy();
  	LCD_PORT_MOD(0x00); lcd_enable();
  	LCD_PORT_MOD(0x0C); lcd_enable();
}
/***************************************************
 *  Turn off display
 ***************************************************/
void lcd_disp_off(void)
{
  	lcd_busy();
  	LCD_PORT_MOD(0x00); lcd_enable();
  	LCD_PORT_MOD(0x08); lcd_enable();
}
/****************************************************
 *  Goto position   (0x00 - 1.row, 1.col)
 *                  (0x40 - 1.row, 2.col)
 ****************************************************/
void lcd_goto(byte pos)
{
	lcd_busy();
	byte temp;
	
  	temp=0x80;
  	temp|=pos;
  	LCD_PORT_MOD((temp >> 4) & 0x0f);
  	lcd_enable();
  	LCD_PORT_MOD(pos & 0x0f);
  	lcd_enable();
}

/****************************************************
 *  Put character on LCD  //edited dberg
 ****************************************************/
void lcd_putch(byte data)
{	
	lcd_busy();
	
	//Send the UPPER nibble first
	LCD_PORT_MOD((data >> 4) & 0x0f);
	SET_RS;
	//CLEAR_RW;
  	lcd_enable();
  	
  	//Now send the LOWER nibble
  	LCD_PORT_MOD(data  & 0x0f);
  	SET_RS;
	//CLEAR_RW;
  	lcd_enable();
  	
  	CLEAR_RS;
}

/***************************************************
 *  Display null terminated string
 ***************************************************/
void lcd_putstr(byte *data)
{
    while(*data) 
    {
        lcd_putch(*data);
        data++;
    }
}

/****************************************************
 *  Display integer
 ****************************************************/
void lcd_printint(unsigned int x)
{
  	char interger[10];
	
	itoa(x,interger,10);
	
	lcd_putstr(interger);
}

/****************************************************
 *  Display byte in hex
 ****************************************************/
void lcd_printhex(byte i)
{
    byte hi,lo;

    hi=i&0xF0;               // High nibble
    hi=hi>>4;
    hi=hi+'0';
    if (hi>'9')
        hi=hi+7;

    lo=(i&0x0F)+'0';         // Low nibble
    if (lo>'9')
        lo=lo+7;

    lcd_putch(hi);
    lcd_putch(lo);
}

